@@ -1,8 +1,13 @@
+// SPDX-License-Identifier: Apache-2.0
+// Copyright Contributors to the OpenTimelineIO project
+
 #pragma once
 
 #include "opentimelineio/serializableObjectWithMetadata.h"
 #include "opentimelineio/version.h"
 
+#include <ImathBox.h>
+
 namespace opentimelineio { namespace OPENTIMELINEIO_VERSION {
 
 class Composition;
@@ -29,6 +34,9 @@ public:
 
     virtual RationalTime duration(ErrorStatus* error_status = nullptr) const;
 
+    virtual optional<Imath::Box2d>
+    available_image_bounds(ErrorStatus* error_status) const;
+
 protected:
     bool        _set_parent(Composition*) noexcept;
     Composable* _highest_ancestor() noexcept;
